﻿namespace Ozeki
{
    class DownloadUI : IOzHfProgressReport
    {
        Dictionary<string, OzHfDownloadProgress> _progressDict = new Dictionary<string, OzHfDownloadProgress> { };
        List<OzHfDownloadEvent> _events = new List<OzHfDownloadEvent>();
        OzHfDownloadEvent? _lastEvent = null;

        Thread _renderProgressThread;
        int _numberOfThreads = 0;
        bool _running = false;

        private readonly object _progressLock = new();
        private readonly object _eventLock = new();

        public DownloadUI()
        {
            _renderProgressThread = new Thread(() =>
            {
                while (_running)
                {
                    Render();
                    Thread.Sleep(100);
                }
            });
        }

        public void Start()
        {
            Console.Clear();
            Console.CursorVisible = false;
            _running = true;       
            _renderProgressThread.Start();
        }

        public void Render()
        {
            var consoleSize = Console.WindowWidth;
            var consoleHeight = Console.WindowHeight;
            var update = false;
            int progressHeight = 0;
            lock (_progressLock)
            {
                var progressBars = _progressDict.Select(x => new ProgressBar(x.Key, x.Value.Status, x.Value.Estimation)).Reverse().ToList();
                if (_numberOfThreads != progressBars.Count)
                {
                    _numberOfThreads = progressBars.Count;
                    update = true;
                }
                Console.SetCursorPosition(0, 0);
                Console.WriteLine("Processes:");

                for (int i = 0; i < progressBars.Count; i++)
                {
                    var progress = i + 1 + " " + progressBars[i].ToString();
                    Console.SetCursorPosition(0, i + 1);
                    Console.Write(progress);
                    for (int j = progress.Length - 1; j < consoleSize - 1; j++)
                    {
                        Console.Write(" ");
                    }
                }
                for (int j = 0; j < consoleSize - 1; j++)
                {
                    Console.Write(" ");
                }
                progressHeight = progressBars.Count;
            }
            lock (_eventLock)
            {
                if (_events.Count < 1)
                {
                    return;
                }
                
                _lastEvent ??= _events.Last();
                if (_lastEvent == _events.Last() && !update)
                {
                    return;
                }
                _lastEvent = _events.Last();

                var eventOutput = _events.TakeLast(10).Reverse().ToList();
                Console.SetCursorPosition(0, 2 + progressHeight);
                Console.Write("Events:");
                for (int j = "Events:".Length - 1; j < consoleSize - 1; j++)
                {
                    Console.Write(" ");
                }
                int i = 0;
                for (; i < eventOutput.Count; i++)
                {
                    Console.SetCursorPosition(0, i + 3 + progressHeight);
                    var message = eventOutput[i].At.ToString("yyyy.MM.dd HH:mm:ss") + ": " + eventOutput[i].Message;
                    var color = eventOutput[i].EventType switch 
                    { 
                        OzHfDownloadEventType.ERROR => ConsoleColor.Red,
                        OzHfDownloadEventType.SUCCESS => ConsoleColor.Green,
                        _ => ConsoleColor.White,
                    };
                    Console.ForegroundColor = color;
                    Console.Write(message);
                    Console.ForegroundColor = ConsoleColor.White;
                    for (int j = message.Length - 1; j < consoleSize - 1; j++)
                    {
                        Console.Write(" ");
                    }                  
                }
            }
        }

        public void Stop()
        {
            if (_running == true) 
            { 
                _running = false;    
                _renderProgressThread.Join();
                Console.Clear();
                Console.CursorVisible = true;
            }
        }

        public void HandleEvent(OzHfDownloadEvent downloadEvent)
        {
            lock (_eventLock)
            {
                _events.Add(downloadEvent);
            }
        }
        public void HandleProgress(OzHfDownloadProgress value)
        {
            lock (_progressLock)
            {
                if (value.Status == OzHfDownloadProgressStatus.OK)
                {
                    var pair = new KeyValuePair<string, OzHfDownloadProgress>(value.Filename, value);
                    _progressDict.Remove(value.Filename);
                }
                else if (value.Estimation == null)
                {
                    return;
                }
                else
                {
                    _progressDict[value.Filename] = value;
                }
            }
        }

        public bool AcceptDownloadDetails(OzHfDownloadDetails downloadDetails, out IEnumerable<OzHfDownloadManifest> filteredManifests)
        {
            var files = downloadDetails.numberOfFiles;
            var total = OzHfDownloadDetails.FileSizeToString(downloadDetails.TotalDownload);
            var avalaible = OzHfDownloadDetails.FileSizeToString(downloadDetails.AvalaibleSpace);
            var remaining = OzHfDownloadDetails.FileSizeToString(downloadDetails.AvalaibleSpace - downloadDetails.TotalDownload + downloadDetails.AlreadyDownloadedSpace);
            var alreadyDownloaded = OzHfDownloadDetails.FileSizeToString(downloadDetails.AlreadyDownloadedSpace);
            IEnumerable<OzHfDownloadManifest> recommendation = [];

            Console.WriteLine($"Number of files to download: {downloadDetails.numberOfFiles}, Total space required: {total}, Avalaible space: {avalaible}, {(downloadDetails.AlreadyDownloadedSpace == 0 ? "" : "Already downloaded: " + alreadyDownloaded + ", ")}Remaining space after download: {remaining}");
            if (downloadDetails.SelectedManifests.Count() == 1)
            {
                var single = downloadDetails.SelectedManifests.First();
                var split = single.FileName.Split('/');
                var location = String.Join('/', split[0..^1]);

                if (split[^1].Contains("-of-"))
                {
                    recommendation = downloadDetails.DownloadManifests.Where(x => String.Join('/', x.FileName.Split('/')[0..^1]) == location && x.FileName.Contains("-of-"));
                    var size = recommendation.Sum(x => x.Size);
                    var partsAvalaible = downloadDetails.AvalaibleSpace - size;
                    Console.WriteLine("Other parts of the selected file can be found. Type 'Parts' to download all other parts");
                    Console.WriteLine($"Number of parts: {recommendation.Count()} Total size of file parts: {OzHfDownloadDetails.FileSizeToString(size)} Remaining space after download: {OzHfDownloadDetails.FileSizeToString(partsAvalaible)}");
                }
            }
            var parts = (recommendation?.Any() ?? false);
            Console.WriteLine($"Do you wish to proceed: [Y]es/[N]o/[S]elect files{(parts ? "/[P]arts}" : "")}?");            
            
            var userinput = Console.ReadLine()?.ToLower();
            filteredManifests = downloadDetails.SelectedManifests;
            if (userinput == null || userinput == "")
            {
                Start();
                return true;
            }
            else if (userinput.StartsWith('n'))
            {
                return false;
            }
            else if (userinput.StartsWith('p') && parts && recommendation != null)
            {
                Start();
                filteredManifests = recommendation;
                return true;
            }
            else if (userinput.StartsWith('y'))
            {
                Start();
                return true;
            }
            else if (userinput.StartsWith('s'))
            {
                Console.Clear();
                OzHfFileSelector checkbox = new OzHfFileSelector(downloadDetails.DownloadManifests, downloadDetails.SelectedManifests.Select(x => x.FileName), downloadDetails.AvalaibleSpace);
                var selectedFileNames = checkbox.Select();
                filteredManifests = downloadDetails.DownloadManifests.Where(x => selectedFileNames.Contains(x.FileName));
                Console.Clear();
                Start();
                return true;
            }
            return false;            
        }
        internal void Initialize()
        {
            Console.WriteLine("Collecting download details...");
        }
    }
}
